@@ -1,5 +1,5 @@
 /* GAIL - The GNOME Accessibility Implementation Library
- * Copyright 2001 Sun Microsystems Inc.
+ * Copyright 2001, 2002, 2003 Sun Microsystems Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -33,8 +33,6 @@ static void                  gail_menu_item_real_initialize
 static gint                  gail_menu_item_get_n_children (AtkObject      *obj);
 static AtkObject*            gail_menu_item_ref_child      (AtkObject      *obj,
                                                             gint           i);
-static AtkStateSet*          gail_menu_item_ref_state_set  (AtkObject      *obj);
-
 static void                  gail_menu_item_finalize       (GObject        *object);
 
 static void                  atk_action_interface_init     (AtkActionIface *iface);
@@ -107,7 +105,6 @@ gail_menu_item_class_init (GailMenuItemClass *klass)
 
   class->get_n_children = gail_menu_item_get_n_children;
   class->ref_child = gail_menu_item_ref_child;
-  class->ref_state_set = gail_menu_item_ref_state_set;
   class->initialize = gail_menu_item_real_initialize;
 
   parent_class = g_type_class_peek_parent (klass);
@@ -147,6 +144,13 @@ gail_menu_item_real_initialize (AtkObject *obj,
     }
   g_object_set_data (G_OBJECT (obj), "atk-component-layer",
                      GINT_TO_POINTER (ATK_LAYER_POPUP));
+
+  if (GTK_IS_TEAROFF_MENU_ITEM (data))
+    obj->role = ATK_ROLE_TEAR_OFF_MENU_ITEM;
+  else if (GTK_IS_SEPARATOR_MENU_ITEM (data))
+    obj->role = ATK_ROLE_SEPARATOR;
+  else
+    obj->role = ATK_ROLE_MENU_ITEM;
 }
 
 static void
@@ -172,16 +176,36 @@ gail_menu_item_new (GtkWidget *widget)
   accessible = ATK_OBJECT (object);
   atk_object_initialize (accessible, widget);
 
-  if (GTK_IS_TEAROFF_MENU_ITEM (widget))
-    accessible->role = ATK_ROLE_TEAR_OFF_MENU_ITEM;
-  else if (GTK_IS_SEPARATOR_MENU_ITEM (widget))
-    accessible->role = ATK_ROLE_SEPARATOR;
-  else
-    accessible->role = ATK_ROLE_MENU_ITEM;
-
   return accessible;
 }
 
+GList *
+get_children (GtkWidget *submenu)
+{
+  GList *children;
+
+  children = gtk_container_get_children (GTK_CONTAINER (submenu));
+  if (g_list_length (children) == 0)
+    {
+      /*
+       * If menu is empty it may be because the menu items are created only 
+       * on demand. For example, in gnome-panel the menu items are created
+       * only when "show" signal is emitted on the menu.
+       *
+       * The following hack forces the menu items to be created.
+       */
+      if (!GTK_WIDGET_VISIBLE (submenu))
+        {
+          GTK_WIDGET_SET_FLAGS (submenu, GTK_VISIBLE);
+          g_signal_emit_by_name (submenu, "show");
+          GTK_WIDGET_UNSET_FLAGS (submenu, GTK_VISIBLE);
+        }
+      g_list_free (children);
+      children = gtk_container_get_children (GTK_CONTAINER (submenu));
+    }
+  return children;
+}
+
 /*
  * If a menu item has a submenu return the items of the submenu as the 
  * accessible children; otherwise expose no accessible children.
@@ -204,7 +228,7 @@ gail_menu_item_get_n_children (AtkObject* obj)
     {
       GList *children;
 
-      children = gtk_container_get_children (GTK_CONTAINER (submenu));
+      children = get_children (submenu);
       count = g_list_length (children);
       g_list_free (children);
     }
@@ -232,7 +256,7 @@ gail_menu_item_ref_child (AtkObject *obj,
       GList *children;
       GList *tmp_list;
 
-      children = gtk_container_get_children (GTK_CONTAINER (submenu));
+      children = get_children (submenu);
       tmp_list = g_list_nth (children, i);
       if (!tmp_list)
         {
@@ -249,30 +273,6 @@ gail_menu_item_ref_child (AtkObject *obj,
   return accessible;
 }
 
-static AtkStateSet*
-gail_menu_item_ref_state_set (AtkObject *obj)
-{
-  AtkStateSet *state_set;
-  GtkWidget *widget;
-  GtkWidget *parent;
-
-  state_set = ATK_OBJECT_CLASS (parent_class)->ref_state_set (obj);
-  widget = GTK_ACCESSIBLE (obj)->widget;
-
-  if (widget == NULL)
-    return state_set;
-
-  if (!GTK_WIDGET_IS_SENSITIVE (widget))
-    atk_state_set_remove_state (state_set, ATK_STATE_SELECTABLE);
-
-  parent = gtk_widget_get_parent (widget);
-  if (!GTK_WIDGET_VISIBLE (parent))
-    atk_state_set_remove_state (state_set, ATK_STATE_VISIBLE);
-  if (!GTK_WIDGET_MAPPED (parent))
-    atk_state_set_remove_state (state_set, ATK_STATE_SHOWING);
-  return state_set;
-}
-
 static void
 atk_action_interface_init (AtkActionIface *iface)
 {
@@ -307,7 +307,7 @@ gail_menu_item_do_action (AtkAction *action,
       if (gail_menu_item->action_idle_handler)
         return FALSE;
       else
-        gail_menu_item->action_idle_handler = gtk_idle_add (idle_do_action, gail_menu_item);
+        gail_menu_item->action_idle_handler = g_idle_add (idle_do_action, gail_menu_item);
       return TRUE;
     }
   else
@@ -560,7 +560,7 @@ gail_menu_item_finalize (GObject *object)
   g_free (menu_item->click_description);
   if (menu_item->action_idle_handler)
     {
-      gtk_idle_remove (menu_item->action_idle_handler);
+      g_source_remove (menu_item->action_idle_handler);
       menu_item->action_idle_handler = 0;
     }
 
